/* XMLOutputStream.java - Copyright (c) 2000 eXcelon Corporation, Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of eXcelon
 * Corporation, Inc. ("Confidential Information").  You shall not disclose such
 * Confidential Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with eXcelon Corporation, Inc.
 *
 * EXCELON CORP MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT. EXCELON SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY 
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR 
 * ITS DERIVATIVES.
 */
package com.exceloncorp.XML;

import java.io.*;
import java.util.*;

/**
 * XMLOutputStream is a class designed to eliminate the errors that can be generatedwhen creating XML documents.
 * It's main aim is to ensure that the document produced is valid, in the sense that the
 * XML is always well formed.
 * 
 * @author D.J.Millman
 * @version 1.0
 */


public class XMLOutputStream
{

   /**
    * Constructor that creates an XMLOutputStream so that all the
    * XML is written to a string that can be used.
    * 
    * @exception java.lang.Exception
    */


   //
   // Constructor(s)
   //
   public XMLOutputStream() throws Exception
   {
      indent = INDENT_INITIAL_VALUE;
      prettyPrint = PRETTY_PRINT_INITIAL_VALUE;
      buffer = new StringBuffer();
      outputString = new StringBuffer();
      _isCommentStarted = false;
      _isStartTagOpen = false;
      _minimizeOutput = false;
      _isStartPIOpen = false;
      _acceptData = false;
      _simpleTag = false;
      _emptyTag = false;
   }

   /**
    * A Constuctor that creates an XMLOuptutStream that writes the
    * information to a file defined by the aFilename parameter.
    * 
    * @param aFilename The name of the file that the XML will be written to.
    * @exception java.lang.Exception
    */
   public XMLOutputStream( String aFilename ) throws Exception
   {
      this();
      if ( aFilename != null )
      {
         _filename = new String(aFilename);
         outputStream = new DataOutputStream(new FileOutputStream(aFilename));  
      }
   }
   
   public void close() throws Exception
   {
        endComment();
        if ( outputStream != null )
        {
            outputStream.close();
            outputStream = null;
        }
   }

   //
   // Document method(s)
   //
   public void startDocument() throws com.exceloncorp.XML.XMLInvalidAttributeNameException,
                                      com.exceloncorp.XML.XMLInvalidAttributeValueException,
                                      com.exceloncorp.XML.XMLDuplicateAttributeException
   {
      endComment();
      closeStartTag();
      writeDocumentHeader();
   }
   public void startDocument( XMLAttributes attributes ) throws com.exceloncorp.XML.XMLInvalidAttributeNameException,
                                                                com.exceloncorp.XML.XMLInvalidAttributeValueException,
                                                                com.exceloncorp.XML.XMLDuplicateAttributeException
   {
      endComment();
      closeStartTag();
      writeDocumentHeader();
   }

   /**
    * endDocument() should be called when all the tags that are
    * required in the document are entered as it will ensure that all
    * the tags that are currently open are closed.
    * 
    * @exception java.lang.Exception
    */

   public void endDocument() throws Exception
   {
      endComment();
      while ( openTags() == true )
      {
         endTag();
      }
   }

   /**
    * Returns the name of the file that the XML has/is being written to
    * 
    * @return The name of the output file.  If the XML is not being written to a file
    * then null is returned.
    */

   public String getFilename()
   {
      return _filename;
   }

   //
   // XML Processing Instruction Information
   //
   public void addPI(String data ) throws Exception
   {
      endComment();
      closeStartTag();
      if ( data != null )
      {
         String xmlData = XMLString.convertStringToXML(data);
         startLine();
	     writeData(XMLProcessingInstructionTag.XML_PROCESSING_INSTRUCTION_START + xmlData);
	     this._isStartPIOpen = true;
      }
   }

   public void addPI(String data, XMLAttributes values ) throws Exception
   {
      endComment();
      closeStartTag();
      if ( data != null )
      {
        addPI(data);
        if ( values != null )
        {
            values.print(this);
        }
      }
   }

   //
   // XML Comment method(s)
   //
   public void startComment()
   {
      startLine();
      this.closeStartTag();
	  writeData(XMLCommentTag.XML_COMMENT_START);
	  _isCommentStarted = true;
	  endLine();
   }
   
   public void addComment( String comment ) throws Exception
   {
      if ( _isCommentStarted == false )
      {
         startComment();
      }
      if ( _isCommentStarted == true )
      {
         if ( comment != null )
         {
            startLine();
            String xmlData = XMLString.convertStringToXML(comment);
            writeData(xmlData);
            endLine();
         }
      }
   }
   
   public void endComment()
   {
      if ( _isCommentStarted == true )
      {
         writeData( XMLCommentTag.XML_COMMENT_END);
         _isCommentStarted = false;
         endLine();
      }
   }

   //
   // XML data output method(s)
   //
   public void writeTagData( String aData )
   {
      if ( _acceptData == true )
      {
         writeData(aData);
         _acceptData = false;
         _simpleTag = true;
         _emptyTag = false;
      }
   }
   public void startTag( String aName ) throws Exception
   {
      endComment();
      this.closeStartTag();
      if ( aName != null )
      {
        try
        {
         mainDocumentStarted = true;
         Vector currentTags = getOpenTags();
         XMLTagName tagValue = new XMLTagName(aName);
         currentTags.addElement(tagValue);
         
         startLine();
	     writeData(MarkUpLanguageTag.MARKUP_START_TAG + aName );
	     _isStartTagOpen = true;
         incrementIndent();
        }
        catch ( Exception e )
        {}
      }
   }

   public void startTag( String aNamespace, String aName )
   {
      endComment();
      this.closeStartTag();
      if (( aName != null ) &&
         ( aNamespace != null ))
      {
        try
        {
         mainDocumentStarted = true;
         Vector currentTags = getOpenTags();
         XMLTagName tagValue = new XMLTagName(aNamespace, aName);
         currentTags.addElement(tagValue);
         
         startLine();
	     writeData(MarkUpLanguageTag.MARKUP_START_TAG + aNamespace + XMLTag.NAMESPACE_SEPARATOR + aName);
         _isStartTagOpen = true;
         incrementIndent();
        }
        catch ( Exception e )
        {}
      }
   }
   
   public void startTag( String aName, XMLAttributes attributes )
   {
      endComment();
      this.closeStartTag();
      if ( aName != null )
      {
         try
         {
            mainDocumentStarted = true;
            Vector currentTags = getOpenTags();
            XMLTagName tagValue = new XMLTagName(aName);
            currentTags.addElement(tagValue);
                 
            startLine();
	        writeData(MarkUpLanguageTag.MARKUP_START_TAG + aName);
            if ( attributes != null )
            {
                attributes.print(this);
            }
            _isStartTagOpen = true;
            incrementIndent();
         }
         catch ( Exception e )
         {}
      }
   }
   
   public void addAttribute( String aName, String aValue )
   {
      if ( this.canWriteAttributes() == true )
      {
         if ( aName != null )
         {
            writeData(" ");
            writeData(aName + "=" +"\"");
            if ( aValue != null )
            {
                writeData(aValue);
            }
            writeData("\"");
         }
      }
   }
   public void endTag()
   {
      this.closeStartTag();
      endComment();
      try
      {
        if ( openTags.size() > 0 )
        {
           XMLTagName tagName = (XMLTagName)openTags.lastElement();
           decrementIndent();
           startLine();
           tagName.print(this);
           if ( openTags.size() > 0 )
           {
              openTags.remove(openTags.size() - 1);
           }
        endLine();
        }
      }
      catch (Exception e )
      {
        System.out.println("End tag exception raised" + e );
      }
   }
   
   public void writeTag( String aNamespace, String aTagName, XMLAttributes aAttributes, String aData )
   {
      this.closeStartTag();
      endComment();
      try
      {
         startLine();
	     writeData("<");
         if ( aNamespace != null )
         {
	     writeData(aNamespace + ":");
         }
	     writeData(aTagName + ">");
         if ( aData != null )
         {
	        writeData(aData);
         }
         writeData("</");
         if ( aNamespace != null )
         {
	        writeData(aNamespace + ":");
         }
	        writeData(aTagName + ">");
	        endLine();
      }
      catch ( Exception e )
      {}
   }

   //
   // XML Format method(s)
   //
   public void enableMinimalOutput()
   {
      _minimizeOutput = true;
   }
   public void disableMinimalOutput()
   {
      _minimizeOutput = false;
   }
   public void toggleMinimalOutput()
   {
      if ( _minimizeOutput == false )
      {
         enableMinimalOutput();
      }
      else
      {
         disableMinimalOutput();
      }
   }
   public boolean isMinimalOutputEnabled()
   {
      return _minimizeOutput;
   }
   public void enablePrettyPrint()
   {
      prettyPrint = true;
   }
   public void disablePrettyPrint()
   {
      prettyPrint = false;
   }
   public boolean isPrettyPrintEnabled()
   {
      return prettyPrint;
   }
   public void togglePrettyPrint()
   {
      if ( prettyPrint == false )
      {
	     disablePrettyPrint();
      }
      else
      {
	     enablePrettyPrint();
      }
   }
   public DataOutputStream getOutputStream()
   {
      return outputStream;
   }

   public String toString()
   {
      String returnValue = null;
      if ( outputString == null )
      {
         returnValue = new String( this.getClass().getName() + " : Filename = " +  _filename);
      }
      else
      {
         returnValue = outputString.toString();
      }
      return returnValue;
   }
   
   protected void add( String aValue )
   {
      try
      {
         if ( aValue != null )
         {
	    writeData(aValue);
         }
      }
      catch ( Exception e )
      {
      }
   }
   
   private void endLine()
   {
      if ( prettyPrint == true )
      {
        try
        {
	       writeData("\n");
        }
        catch (Exception e )
        {}
      }
   }
   private void startLine()
   {
      if ( prettyPrint == true )
      {
         writeIndent();
      }
   }
   
   private void incrementIndent()
   {
      indent++;
      buffer.append(INDENT_VALUE);
   }
   
   private void decrementIndent()
   {
      if ( indent > INDENT_INITIAL_VALUE )
      {
        indent--;
        if ( indent > 0 )
        {
           buffer.delete(((indent * INDENT_SIZE) - 1),(buffer.length()-1));
        }
        else
        {
           buffer.delete(0,(buffer.length()-1));
        }
      }
   }
   
   private void writeIndent()
   {
      try
      {
        if ( indent > 0 )
        {
	       writeData(buffer.substring(INITIAL_STRING_BUFFER_INDEX));
        }
      }
      catch ( Exception e )
      {
      }
   }
   
   private Vector getOpenTags()
   {
      if ( openTags == null )
      {
        openTags = new Vector();
      }
      return openTags;
   }
   
   private boolean openTags()
   {
      boolean returnValue = false;
      if ( openTags != null )
      {
        if ( openTags.size() > NO_OPEN_TAG_CARDINALITY )
        {
            returnValue = true;
        }
      }
      return returnValue;
   }

   private void writeData( String data )
   {
      if ( outputStream != null )
      {
         try
	 {
            outputStream.writeBytes(data);
	 }
	 catch (IOException e )
	 {
	 }
      }
      if ( outputString != null )
      {
         outputString.append(data);
      }
   }
   
   private void closeComment()
   {
   }
   
   private void closeStartTag()
   {
      if ( _isStartPIOpen == true )
      {
         writeData(XMLProcessingInstructionTag.XML_PROCESSING_INSTRUCTION_END);
         _isStartPIOpen = false;
         endLine();
      }
      else if ( _isStartTagOpen == true )
      {
         writeData(MarkUpLanguageTag.MARKUP_END_TAG);
         _isStartTagOpen = false;
         endLine();
      }
   }
   
   private boolean canWriteAttributes()
   {
      boolean returnValue = false;
      if ( _isStartPIOpen == true )
      {
         returnValue = true;
      }
      else if ( _isStartTagOpen )
      {
         returnValue = true;
      }
      return returnValue;
   }
   
   private void writeDocumentHeader()
   {
      try
      {
         this.addPI("xml");
         this.addAttribute("version", "1.0");
      }
      catch ( Exception e )
      {
      }
   }

   //
   // Attribute(s)
   //
   private DataOutputStream outputStream;   // The file stream that the XML is being written to
   private StringBuffer     outputString;   // The string that the XML is being written to.
   private boolean          prettyPrint;    // True pretty print is on, false it is off.
   private int              indent;         // Number of indents
   private StringBuffer     buffer;         // Indent Buffer
   private Vector           openTags;       // list of tags currently open
   private boolean          mainDocumentStarted;  // has a document been started
   private String           _filename;            // name of the file that is being written to.
   private boolean          _isCommentStarted;    // true if a comment block has been started
   private boolean          _isStartTagOpen;      // true is the start Tag has not been closed (i.e. attributes can be added);
   private boolean          _isStartPIOpen;
   private boolean          _minimizeOutput;      // if true then only the smallest amount of data will be entered into the stream.
   private boolean          _acceptData;          // if true then the client can write data to the output stream.
   private boolean          _simpleTag;           // if true then the current tag is simple otherwise it is not.
   private boolean          _emptyTag;            // Used in combination with _simpleTag, if true then no data is stored/
   
   //
   // Constant(s)
   //
   private static final int     INDENT_INITIAL_VALUE         = 0;
   private static final String  INDENT_VALUE                 = "   ";
   private static final int     INDENT_SIZE                  = INDENT_VALUE.length();
   private static final boolean PRETTY_PRINT_INITIAL_VALUE   = true;
   private static final int     INITIAL_STRING_BUFFER_INDEX  = 0;
   private static final int     NO_OPEN_TAG_CARDINALITY      = 0;
   
   
   public static void main( String[] args )
   {
    try
    {
      XMLOutputStream doc = new XMLOutputStream();
      //doc.disablePrettyPrint();
      doc.enableMinimalOutput();
      doc.startDocument();
      doc.addAttribute("encoding","abcderfg");
      doc.startComment();
      doc.addComment("Name = kjasdhflkjhadslfkjhlaksdhflkjhsalkdfjhlksadjflkasjhdflkjdf");
      doc.addComment("asdh =sadfnhadslkfjadksljflkadshflkh");
      doc.addPI("pitest");
      doc.addAttribute("myAttrib","value");
      doc.addAttribute("attrib2","");
      doc.startTag("abc");
      doc.addAttribute("abcdef","asdj");
      doc.startTag("bcd");
      doc.writeTag("abc","myTag",null,"Data");
      doc.startTag("cde");
      doc.endDocument();
      System.out.println(doc.toString());
    }
    catch (Exception e )
    {
        System.out.println(e);
    }
   }
   
}
