/*
 * XMLOutputStream.java - Copyright (c) 2000 eXcelon Corporation, Inc. All
 * Rights Reserved. This software is the confidential and proprietary
 * information of eXcelon Corporation, Inc. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into with
 * eXcelon Corporation, Inc. EXCELON CORP MAKES NO REPRESENTATIONS OR WARRANTIES
 * ABOUT THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. EXCELON SHALL NOT BE LIABLE FOR ANY
 * DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 */
package com.sonicsw.sts.xmlstream;

import java.io.DataOutputStream;

import org.w3c.dom.Node;

/**
 * XMLOutputStream is a class designed to eliminate the errors that can be
 * generatedwhen creating XML documents. It's main aim is to ensure that the
 * document produced is valid, in the sense that the XML is always well formed.
 * All reserved characters such as < are automatically translated prior to
 * placing in the stream.
 * 
 * @author D.J.Millman
 * @version 1.0
 */

public interface XMLOutputStream {
   public void close() throws Exception;

   /**
    * Start document should be the first method called, this starts off by
    * reseting all the variables in the actual implementation and defining the
    * starting tags for the document structure, i.e. <? xml version="1.0". Note
    * the start tag will not be closed until the next tag is started or the
    * document is finished, this allows the addAttribute() method to be called
    * to add other attributes to the document definition.
    * 
    * @exception com.sonicsw.sts.xmlstream.XMLInvalidAttributeNameException
    * @exception com.sonicsw.sts.xmlstream.XMLInvalidAttributeValueException
    * @exception com.sonicsw.sts.xmlstream.XMLDuplicateAttributeException
    */
   //
   // Document method(s)
   //
   public void startDocument()
         throws com.sonicsw.sts.xmlstream.XMLInvalidAttributeNameException,
         com.sonicsw.sts.xmlstream.XMLInvalidAttributeValueException,
         com.sonicsw.sts.xmlstream.XMLDuplicateAttributeException;

   /**
    * Start document should be the first method called, this starts off by
    * reseting all the variables in the actual implementation and defining the
    * starting tags for the document structure, i.e. <? xml version="1.0" ?>.
    * The start tag will be closed at the end of this call and therefore all the
    * attributes have to be passed in the attributes parameter.
    * 
    * @param attributes
    *           The definition of all the attributes for the document header
    *           tag.
    * @exception com.sonicsw.sts.xmlstream.XMLInvalidAttributeNameException
    * @exception com.sonicsw.sts.xmlstream.XMLInvalidAttributeValueException
    * @exception com.sonicsw.sts.xmlstream.XMLDuplicateAttributeException
    */
   public void startDocument(XMLAttributes attributes)
         throws com.sonicsw.sts.xmlstream.XMLInvalidAttributeNameException,
         com.sonicsw.sts.xmlstream.XMLInvalidAttributeValueException,
         com.sonicsw.sts.xmlstream.XMLDuplicateAttributeException;

   /**
    * endDocument() should be called when all the tags that are required in the
    * document are entered as it will ensure that all the tags that are
    * currently open are closed.
    * 
    * @exception java.lang.Exception
    */

   public void endDocument() throws Exception;

   /**
    * closeTags( XMLTagMarker aMarker ) closes all the tags that are open since
    * the last
    */
   public void closeTags(XMLTagMarker aMarker);

   /**
    * This method returns an instance of XMLTagMarker that is a reference to the
    * current part of the document. This marker can be used with the closeTags()
    * method to close all the XML tags that have been opened since the marker
    * was retrieved.
    * 
    * @return A new instance of com.exceloncorp.XML.XMLTagMarker.
    */

   public XMLTagMarker getTagMarker();

   /**
    * Returns the name of the file that the XML has/is being written to
    * 
    * @return The name of the output file. If the XML is not being written to a
    *         file then null is returned.
    */

   public String getFilename();

   /**
    * Add a process instruction to the output stream
    * 
    * @param data
    * @exception java.lang.Exception
    */

   //
   // XML Processing Instruction Information
   //
   /**
    * Create a processing instruction in the XML output and place it in the
    * stream, the content of the processing instruction will be the value of
    * data.
    */
   public void addPI(String data) throws Exception;

   public void addPI(String data, XMLAttributes values) throws Exception;

   /**
    * Start a comment area in the XML. The comment area will be terminated by
    * either calling endComment() or starting a new tag.
    */

   //
   // XML Comment method(s)
   //
   /**
    * Start a new comment in the output stream
    */
   public void startComment();

   /**
    * Add a comment or extend a new comment
    * 
    * @param comment
    *           The comment to be placed in the comment block.
    * @exception java.lang.Exception
    */

   public void addComment(String comment) throws Exception;

   /**
    * Finish the current comment
    */
   public void endComment();

   //
   // DTD specific method(s)
   //
   public void writePublicDTD(String aRootElement, String aName, String aURI);

   public void writeSystemDTD(String aRootElement, String aURI);

   //
   // XML data output method(s)
   //
   public void writeTagData(String aData);

   /**
    * Start a new XML tag
    * 
    * @param aName
    *           the name of the tag to start
    */
   public void startTag(String aName) throws Exception;

   /**
    * Start a new tag with a specific namespace. When a tag has been started the
    * any number of attributes can be written with the addAttribute() methods.
    * 
    * @param aNamespace
    *           the namespace that is associated with the tag
    * @param aName
    *           the name of the tag
    */
   public void startTag(String aNamespace, String aName);

   /**
    * Start a new tag and also pass some attributes to the value.
    * 
    * @param aName
    *           the name of the tag to start
    * @param attributes
    *           the attributes that are placed in the tag
    */
   public void startTag(String aName, XMLAttributes attributes);

   /**
    * Add an attribute if a tag is currently open
    * 
    * @param argName
    *           the name of the attribute to add
    * @param argValue
    *           the value of the attribute to be added
    */
   public void addAttribute(String argName, String argValue);

   /**
    * Add an attribute to the current open tag
    * 
    * @param argNamespace
    *           the namespace of the tag
    * @param argName
    *           the name of the attribute to add
    * @param argValue
    *           the value of the attribute, if null then no data is written
    */
   public void addAttribute(String argNamespace, String argName, String argValue);

   /**
    * End the current tag that is open
    */
   public void endTag();

   /**
    * Write a tag in the form <aNamespace:aTagName aAttributes>aData
    * </aNamespace:aTagName>
    * 
    * @param argNamespace
    *           the namespace used for this tag, or null if no namespace is used
    * @param argAttributes
    *           the set of attributes to be placed in the tag, or null for no
    *           attributes
    * @param argData
    *           the data to be placed in the tag, or null if no data is written.
    */
   public void writeTag(String argNamespace, String argTagName,
         XMLAttributes argAttributes, String argData);

   public void writeTag(String argNamespace, String argTagName,
         XMLAttributes argAttributes, String argData,
         boolean argWriteIfNullString);

   /**
    * write a tag with int as the data in the form <aNamespace:aTagName
    * aAttributes>aData </aNamespace:aTagName>
    * 
    * @param argNamespace
    *           the namespace of the tag or null if no namespace is used
    * @param argTagName
    *           the name of the tag, this cannot be null and should be a valid
    *           tag name
    * @param argAttributes
    *           the attributes that are to be placed in the tag
    * @param argData
    *           the integer data to be placed in the data section
    */
   public void writeTag(String argNamespace, String argTagName,
         XMLAttributes argAttributes, int argData);

   public void writeTag(String aNamespace, String aTagName,
         XMLAttributes aAttributes, long aData);

   public void writeTag(String aNamespace, String aTagName,
         XMLAttributes aAttributes, boolean aData);

   /**
    * Allow a snippet of XML to be added to the current stream. Note that any
    * leading document information will be lost.
    * 
    * @param argXML
    *           the XML data to be added to the stream
    */
   public void addXMLData(String argXML) throws Exception;

   public void addXMLData(String argXML, boolean argParse) throws Exception;
   
   
   public void addXMLData( Node argNode ) throws Exception;

   //
   // void writeCData( String aData )
   //
   // This method writes out the character data that will be converted into XML,
   // i.e. find
   // and replace all '<' characters etc.
   //
   public void writeCData(String aData);

   public void writePCData(String aData);

   //
   // void writeEmptyTag( String, String )
   //
   // Writes a tag that may contain attributes to the stream however it will not
   // contain any data.
   //
   public void writeEmptyTag(String aNamespace, String aTagName);

   public void add(String aValue);

   //
   // XML Format method(s)
   //
   public void enableMinimalOutput();

   public void disableMinimalOutput();

   public void toggleMinimalOutput();

   /**
    * @return Returns 'true' if the stream will not add any formatting text
    *         (,i.e. pretty-print) has been turned off, otherwise 'false' will
    *         be returned.
    */
   public boolean isMinimalOutputEnabled();

   /**
    * enablePrettyPrint() will ensure that formatting of the outputted XML is
    * automatically created, thereby making the output more readable. Note that
    * the pretty print functionality can be turned on/off in mid-output so as to
    * highlight various areas. The downside of setting this option is that the
    * XML created will be larger as it will contain formatting spaces.
    * 
    * @since 1.0
    */
   public void enablePrettyPrint();

   /**
    * Disable all the pretty print options.
    * 
    * @since 1.0
    */
   public void disablePrettyPrint();

   /**
    * A predicate to determine if the output is using the pretty-print option.
    * 
    * @return Returns 'true' if the current output option has enable
    *         pretty-print, otherwise 'false' is returned.
    * @since 1.0
    */
   public boolean isPrettyPrintEnabled();

   public void togglePrettyPrint();

   public DataOutputStream getOutputStream();

   /**
    * @return Returns a String containing the outputed XML.
    * @since 1.0
    */
   public String toString();

}