package com.sonicsw.sts.xmlstream;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;

/**
 * This class is responsible for converting XML data into method calls on the
 * XMLOutputStream
 */

public class XMLStreamContentHandlerImpl implements ContentHandler {

   /**
    * Create an instance of the ContentHandler that maps to a XMLOutputStream,
    * the stream is where the contents of the document read by the content
    * handler will be written to.
    * 
    * @param argStream
    *           the stream that the XML data will be written to.
    */
   protected XMLStreamContentHandlerImpl(XMLOutputStream argStream) {
      stream_ = argStream;
   }

   public void setDocumentLocator(Locator locator) {
      isStartElement_ = false;
   }

   public void skippedEntity(String name) {
      isStartElement_ = false;
   }

   public void startDocument() throws SAXException {
      isStartElement_ = false;
   }

   public void endDocument() throws SAXException {
      isStartElement_ = false;
   }

   public void characters(char[] ch, int start, int length) {
      if ( isStartElement_ == true )
      {
         currentData_ = new String(ch, start, length);
      }
      isStartElement_ = false;
   }

   public void ignorableWhitespace(char[] ch, int start, int length) {
      isStartElement_ = false;
   }

   public void processingInstruction(String target, String data) {
      isStartElement_ = false;
   }

   /**
    * StartElement() is called everytime that a new tag is detected in the XML
    * that is being parsed.
    * 
    * @param argNamespaceURI
    * @param argLocalName
    * @param argRawName
    * @param argAttributes
    */
   public void startElement(String argNamespaceURI, String argLocalName,
         String argRawName, Attributes argAttributes) {
      if ( isStartElement_ == true )
      {
         if ( this.isATagDefined() == true )
         {
            getStream().startTag(this.getNamespace(), this.getTagName());
            writeAttributes();
            resetValues();
         }
      }
      isStartElement_ = true;
   }

   /**
    * This method is called when the parser detectes the end of a particular XML
    * element, such as a closing tag.
    * 
    * @param argNamespaceURI
    *           ignored in this method, see startElement()
    * @param argLocalName
    *           ignored in this method, see startElement()
    * @param argRawName
    *           ignored in this method, see srartElement()
    */
   public void endElement(String argNamespaceURI, String argLocalName,
         String argRawName) {
      if ( currentData_ != null )
      {
         System.out.println(currentData_);
         currentData_ = null;
      }
      isStartElement_ = false;
      getStream().endTag(); // End the current tag
   }

   public void startPrefixMapping(String prefix, String uri) {
      isStartElement_ = false;
   }

   public void endPrefixMapping(String prefix) {
      isStartElement_ = false;
   }

   /**
    * Return the output stream where the XML will be written to.
    * 
    * @return XMLOutputStream instance that holds is where the XML is written
    *         out to
    */
   private XMLOutputStream getStream() {
      return stream_;
   }

   private void writeData() {
      if ( currentData_ != null )
      {
         if ( currentData_.length() > 0 )
         {
            getStream().writeTag(currentNamespace_, currentTagName_, null,
                  currentData_);
         }
      }
   }

   /**
    * Reset all the values that have been read so that new ones can be read in
    * correctly
    */
   private void resetValues() {
      currentNamespace_ = null;
      currentTagName_ = null;
      currentAttributes_ = null;
      currentData_ = null;
   }

   /**
    * Predicate to determine if this class contains information about a tag at
    * the current time
    * 
    * @return true if information about a tag is currently held.
    */
   private boolean isATagDefined() {
      boolean returnValue = false;
      if ( currentTagName_ != null )
      {
         returnValue = true;
      }
      return returnValue;
   }

   /**
    * Return the current namespace
    * 
    * @return the current namespace or null if it is not specified
    */
   private String getNamespace() {
      return currentNamespace_;
   }

   /**
    * Return the name of the tag that is currently being processed
    * 
    * @return the name of the tag or null if one is not being worked on
    */
   private String getTagName() {
      return currentTagName_;
   }

   /**
    * Retrieve the attributes that are currently associated with the stream, or
    * null if no attributes are associated.
    * 
    * @return all attributes associated with the current tag
    */
   private Attributes getAttributes() {
      return currentAttributes_;
   }

   /**
    * Return the attributes currently read from the XML stream in the internal
    * format XMLAttribute, if no attributes are found then null is returned.
    * 
    * @return the current attributes or null if none exist.
    */
   private XMLAttributes getXMLAttributes() {
      XMLAttributes returnValue = null;
      if ( (currentAttributes_ != null)
            && (currentAttributes_.getLength() > NO_ATTRIBUTES_LENGTH) )
      {
         returnValue = new XMLAttributes();
         for (int currentAttributeIndex = 0; currentAttributeIndex < getAttributes()
               .getLength(); currentAttributeIndex++)
         {
            returnValue.addAttribute(getAttributes().getLocalName(
                  currentAttributeIndex), getAttributes().getValue(
                  currentAttributeIndex));
         }
      }
      return returnValue;
   }

   /**
    * Write the current attributes to the output stream specified
    */
   private void writeAttributes() {
      Attributes attrs = getAttributes();
      if ( (attrs != null) && (attrs.getLength() > NO_ATTRIBUTES_LENGTH) )
      {
         for (int currentAttributeIndex = 0; currentAttributeIndex < attrs
               .getLength(); currentAttributeIndex++)
         {
            getStream().addAttribute(attrs.getLocalName(currentAttributeIndex),
                  attrs.getValue(currentAttributeIndex));
         }
      }
   }

   //
   // Attribute(s)
   private boolean          isStartElement_      = false;

   private String           currentNamespace_;

   private String           currentTagName_;

   private Attributes       currentAttributes_;

   private String           currentData_;

   private XMLOutputStream  stream_;

   //
   // Constant(s)
   //
   private static final int NO_ATTRIBUTES_LENGTH = 0;

}